SET NOCOUNT ON;
go
USE master;
go
-- Drop test database if it exists
IF EXISTS
 (SELECT *
  FROM   sys.databases
  WHERE  name = N'ImEx')
BEGIN;
  -- Force me out if I forgot to close another window :-)
  ALTER DATABASE ImEx
  SET            SINGLE_USER
  WITH  ROLLBACK IMMEDIATE;
  -- Now drop it
  DROP  DATABASE ImEx;
END;
go
-- Create a database for our tests.
-- For optimal performance:
-- a) Put log file and data file on seperate drives, if available (not on my laptop, unfortunately)
-- b) Make sure data and log are large enough, so that autogrow won't kick in
--    (Dut to heavy updates in some algorithms, log needs to be much bigger than data!)
-- c) Use binary collation for fastest string comparison
CREATE DATABASE ImEx
       ON      (NAME = IMEX_Data, FILENAME = N'C:\Program Files\Microsoft SQL Server\MSSQL.1\MSSQL\DATA\ImEx_Data.MDF', SIZE =  50MB, FILEGROWTH = 0)
       LOG ON  (NAME = IMEX_Log,  FILENAME = N'C:\Program Files\Microsoft SQL Server\MSSQL.1\MSSQL\DATA\ImEx_Log.LDF',  SIZE = 250MB, FILEGROWTH = 0)
       COLLATE  Latin1_General_BIN;
-- Don't need full recovery here.
ALTER DATABASE ImEx
  SET RECOVERY SIMPLE;
go

USE ImEx;
go

-- Create table for the sessions.
-- NOTE: For performance reasons, the CHECK constraints are commented out.
--       Since their columns are populated from columns in Regsitrations (where the
--       equivalent constraints are enabled), they'll never be violated anyway.
CREATE TABLE dbo.Sessions
            (Year smallint NOT NULL
            ,Quarter tinyint NOT NULL
            ,SessionNo smallint NOT NULL
            ,SpaceLeft smallint NOT NULL     -- could be calculated each time it's needed,
                                             --  but that would seriously hurt performance.
            ,CONSTRAINT PK_Sessions
                 PRIMARY KEY (Year, Quarter, SessionNo)
--          ,CONSTRAINT CK_Sessions_Year
--               CHECK (Year BETWEEN 1995 AND 2020)
--          ,CONSTRAINT CK_Sessions_Quarter
--               CHECK (Quarter IN (1, 2, 3, 4))
            );
go

-- Create the table for the registrations.
-- NOTE: In a real situation, SubjCode would refer to a Subjects table.
--       Since we don't have this table in our example, I commented the FOREIGN KEY constraint.
-- NOTE: FOREIGN KEY between Registrations and Sessions is not created.
--       This gains an enormous amount of performance, and allows us to create the reference
--       to a session BEFORE actually inserting the row in the Sessions table.
--       Constraint should be added ("WITH CHECK" !!) when done!
CREATE TABLE dbo.Registrations
            (Year smallint NOT NULL
            ,Quarter tinyint NOT NULL
            ,SubjCode char(3) NOT NULL
            ,NumCandidates smallint NOT NULL
            ,SessionNo smallint DEFAULT NULL
            ,CONSTRAINT PK_Registrations
                 PRIMARY KEY (Year, Quarter, SubjCode)
--          ,CONSTRAINT FK_Registrations_Subjects
--               FOREIGN KEY (SubjCode) REFERENCES Subjects (SubjCode)
--          ,CONSTRAINT FK_Registrations_Session
--               FOREIGN KEY (Year, Quarter, SessionNo)
--               REFERENCES Sessions (Year, Quarter, SessionNo)
            ,CONSTRAINT CK_Registrations_Year
                 CHECK (Year BETWEEN 1995 AND 2020)  -- Sanity check
            ,CONSTRAINT CK_Registrations_Quarter
                 CHECK (Quarter IN (1, 2, 3, 4))
            ,CONSTRAINT CK_Registrations_NumCandidates
                 CHECK (NumCandidates > 0 AND NumCandidates <= 100)
            );
go

-- We'll need a table of numbers for one of the solutions.
-- Most databases already have one (and others should!) - see http://www.aspfaq.com/show.asp?id=2516
-- A range of 1 - 999 is more than enough in this case, so I'll use smallint.
CREATE TABLE dbo.Numbers 
            (Number smallint NOT NULL
            ,CONSTRAINT PK_Numbers
                 PRIMARY KEY (Number)
            );
WITH Digits(Digit)
AS (SELECT 0 UNION ALL SELECT 1 UNION ALL SELECT 2 UNION ALL
    SELECT 3 UNION ALL SELECT 4 UNION ALL SELECT 5 UNION ALL
    SELECT 6 UNION ALL SELECT 7 UNION ALL SELECT 8 UNION ALL SELECT 9)
INSERT INTO dbo.Numbers (Number)
SELECT      h.Digit * 100 + t.Digit * 10 + s.Digit
FROM        Digits AS h
CROSS JOIN  Digits AS t
CROSS JOIN  Digits AS s
WHERE       h.Digit * 100 + t.Digit * 10 + s.Digit > 0;
go
